import numpy


def bit_vector_to_int_large(v: numpy.ndarray) -> int:
    """
    Transform a numpy vector representing a sequence of binary bits [0 | >0]
    into an integer representation.

    This function is the special form that can handle very large integers
    (>64bit).

    :param v: 1D Vector of bits

    :return: Integer equivalent

    """
    c = 0
    for b in v:
        c = (c << 1) + int(b)
    return c


def int_to_bit_vector_large(integer: int, bits: int = 0) -> numpy.ndarray:
    """
    Transform integer into a bit vector, optionally of a specific length.

    This function is the special form that can handle very large integers
    (>64bit).

    :raises ValueError: If ``bits`` specified is smaller than the required bits
        to represent the given ``integer`` value.

    :param integer: integer to convert
    :param bits: Optional fixed number of bits that should be represented by the
        vector.

    :return: Bit vector as numpy array (big endian).

    """
    # Can't use math version because floating-point precision runs out after
    # about 2^48
    # -2 to remove length of '0b' string prefix
    size = len(bin(integer)) - 2

    if bits and (bits - size) < 0:
        raise ValueError(
            "%d bits too small to represent integer value %d." % (bits, integer)
        )

    # Converting integer to array
    v = numpy.zeros(bits or size, numpy.bool_)
    for i in range(0, size):
        v[-(i + 1)] = integer & 1
        integer >>= 1

    return v



from math import pi

import numpy as np
from scipy.spatial.distance import cdist


def histogram_intersection_distance(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    """
    Compute the histogram intersection distance between given histogram
    vectors or matrices ``a`` and ``b``, returning a value between ``0.0`` and
    ``1.0``. This is the inverse of intersection similarity, whereby a distance
    of  ``0.0`` means full intersection and ``1.0`` means no intersection.

    This implements non-branching formula for efficient computation.

    Input vectors ``a`` and ``b`` may be of 1 or 2 dimensions. Depending on the
    values of ``a`` and ``b``, different things may occur:

    * If both ``a`` and ``b`` are 2D matrices, they're shape must be congruent
      and the result will be an vector of distances between parallel rows.

    * If either is a 1D vector and the other is a 2D matrix, a vector of
      distances is returned between the 1D vector and each row of the 2D matrix.

    * If both ``a`` and ``b`` are 1D vectors, a floating-point scalar distance
      is returned that is the histogram intersection distance between the input
      vectors.

    :param a: Histogram or array of histograms ``a``
    :param b: Histogram or array of histograms ``b``

    :return: Float or array of float distance (inverse percent intersection).
    """
    # TODO: input value checks?
    # Which axis to sum on. If there is a matrix involved, its along column,
    # but if its just two arrays its along the row.
    #
    # - The following are noticeably slower:
    #       sum_axis = not (a.ndim == 1 and b.ndim == 1)
    #       sum_axis = (a.ndim > 1) | (b.ndim > 1)
    sum_axis = 1
    if a.ndim == 1 and b.ndim == 1:
        sum_axis = 0
    # TODO(john.moeller): Assuming each sums to 1, this can be sped up
    #   return np.abs(np.subtract(i, j)).sum(sum_axis) * 0.5
    return 1. - ((np.add(a, b) - np.abs(np.subtract(a, b))).sum(sum_axis) * 0.5)


def histogram_intersection_distance_fast(i: np.ndarray, j: np.ndarray) -> float:
    """
    Compute the histogram intersection percent relation between given 1D
    histogram vectors ``a`` and ``b``, returning a value between 0.0 and 1.0.
    0.0 means full intersection, and 1.0 means no intersection.

    This implements non-branching formula for efficient computation.

    Use of this implementations is faster when the input will only be 1D
    vectors.

    PENDING DEPRECATION: This function doesn't time much, if any, better than
        ``histogram_intersection_distance`` (via ipython %timeit using random
        input).

    :param i: Histogram ``i``
    :param j: Histogram ``j``

    :return: Float inverse percent intersection amount.

    """
    return 1.0 - ((i + j - np.abs(i - j)).sum() * 0.5)


def euclidean_distance(i: np.ndarray, j: np.ndarray) -> np.ndarray:
    """
    Compute euclidean distance between two N-dimensional point vectors.

    :param i: Vector i
    :param j: Vector j

    :return: Float distance.

    """
    sum_axis = 1
    if i.ndim == 1 and j.ndim == 1:
        sum_axis = 0
    return np.sqrt(np.square(i - j).sum(sum_axis))


def cosine_similarity(i: np.ndarray, j: np.ndarray) -> np.ndarray:
    """
    Angular similarity between vectors i and j. Results in a value between 1,
    where i and j are exactly the same, to -1, meaning exactly opposite. 0
    indicates orthogonality. Negative values will only be returned if input
    vectors can have negative values.

    See: http://en.wikipedia.org/wiki/Cosine_similarity

    :param i: Vector i
    :param j: Vector j

    :return: Float or array of cosine similarity.
    """
    # Only handling 1D-vectors for i
    assert i.ndim == 1

    i = i.reshape(1, -1)
    if j.ndim == 1:
        j = j.reshape(1, -1)

    # Cosine similarity function returns matrix in shape of (i_N_samples, j_N_samples)
    cosine_s = 1 - cdist(i, j, metric='cosine')[0]

    # Return a float if both i and j only have one sample
    if cosine_s.size == 1:
        return cosine_s[0]
    else:
        return cosine_s


def cosine_distance(i: np.ndarray, j: np.ndarray, pos_vectors: bool = True) -> np.ndarray:
    """
    Cosine similarity converted into angular distance.

    See: https://en.wikipedia.org/wiki/Cosine_similarity, section
    "Angular distance and similarity".

    :param i: Vector i
    :param j: Vector j
    :param pos_vectors: If we expect vector elements to always be positive.
        Default value is True (common case).

    :return: Float or array of cosine distance between [0, 1] range.
    """
    # limit to between -1 and 1
    sim = np.maximum(np.minimum(cosine_similarity(i, j), 1), -1)

    return (1 + bool(pos_vectors)) * np.arccos(sim) / pi


def hamming_distance(i: int, j: int) -> int:
    """
    Return the hamming distance between the two given pythonic integers, or the
    number of places where the bits differ.

    **Note:** *We say "pythonic" integer here because this function has no cap
    on the number of bits used to represent said integer. This function will
    execute correctly regardless whether i/j is 32 bits or 512 bits, etc."

    :param i: First integer.
    :param j: Second integer.

    :return: Integer hamming distance between the two values.
    """
    # TODO: Find something better than this?
    return bin(i ^ j).count('1')



from .interfaces.nearest_neighbor_index import NearestNeighborsIndex  # noqa: F401
from .interfaces.lsh_functor import LshFunctor  # noqa: F401
from .interfaces.hash_index import HashIndex  # noqa: F401



import threading
import time
from typing import Callable


class ProgressReporter:
    """
    Helper utility for reporting the state of a loop and the rate at which
    looping is occurring based on lapsed wall-time and a given reporting
    interval.

    Includes optional methods that are thread-safe.

    TODO: Add parameter for an optionally known total number of increments.
    """

    def __init__(
        self,
        log_func: Callable[[str], None],
        interval: float,
        what_per_second: str = "Loops"
    ):
        """
        Initialize this reporter.

        :param log_func: Logging function to use.
        :param interval: Time interval to perform reporting in seconds.  If no
            reporting during incrementation should occur, infinity should be
            passed.
        :param str what_per_second:
            String label about what is happening or being iterated over per
            second. The provided string should make sense when followed by
            " per second ...".

        """
        self.log_func = log_func
        self.interval = float(interval)
        self.what_per_second = what_per_second

        self.lock = threading.RLock()
        # c_last : Increment count at the time of the last report. Updated after
        #          report in ``increment_report``.
        # c      : Current Increment count, updated in ``increment_report``.
        # c_delta: Delta between the increment current and previous count at the
        #          time of the last report. Updated at the time of reporting in
        #          ``increment_report``.
        self.c_last = self.c = self.c_delta = 0
        # t_last : Time of the last report. Updated after report in
        #          ``increment_report``.
        # t      : Current time, Updated in ``increment_report``
        # t_delta: Delta between current time and the time of the last report.
        #          Updated in ``increment_report``.
        self.t_last = self.t = self.t_delta = self.t_start = 0.0

        self.started = False

    def start(self) -> "ProgressReporter":
        """ Start the timing state of this reporter.

        Repeated calls to this method resets the state of the reporting for
        multiple uses.

        This method is thread-safe.

        :returns: Self
        :rtype: ProgressReporter

        """
        with self.lock:
            self.started = True
            self.c_last = self.c = self.c_delta = 0
            self.t_last = self.t = self.t_start = time.time()
            self.t_delta = 0.0
        return self

    def increment_report(self) -> None:
        """
        Increment counter and time since last report, reporting if delta exceeds
        the set reporting interval period.
        """
        if not self.started:
            raise RuntimeError("Reporter needs to be started first.")
        self.c += 1
        self.c_delta = self.c - self.c_last
        self.t = time.time()
        self.t_delta = self.t - self.t_last
        # Only report if its been ``interval`` seconds since the last
        # report.
        if self.t_delta >= self.interval:
            self.report()
            self.t_last = self.t
            self.c_last = self.c

    def increment_report_threadsafe(self) -> None:
        """
        The same as ``increment_report`` but additionally acquires a lock on
        resources first for thread-safety.

        This version of the method is a little more costly due to the lock
        acquisition.
        """
        with self.lock:
            self.increment_report()

    def report(self) -> None:
        """
        Report the current state.

        Does nothing if no increments have occurred yet.
        """
        if not self.started:
            raise RuntimeError("Reporter needs to be started first.")
        # divide-by-zero safeguard
        if self.t_delta > 0 and (self.t - self.t_start) > 0:
            self.log_func("%s per second %f (avg %f) "
                          "(%d current interval / %d total)"
                          % (self.what_per_second,
                             self.c_delta / self.t_delta,
                             self.c / (self.t - self.t_start),
                             self.c_delta,
                             self.c))

    def report_threadsafe(self) -> None:
        """
        The same as ``report`` but additionally acquires a lock on
        resources first for thread-safety.

        This version of the method is a little more costly due to the lock
        acquisition.
        """
        with self.lock:
            self.report()



"""
Interface and plugin getter for LSH algorithm hash generation functors.
"""
import abc

import numpy as np

from smqtk_core import Configurable, Pluggable


class LshFunctor (Configurable, Pluggable):
    """
    Locality-sensitive hashing functor interface.

    The aim of such a function is to be able to generate hash codes
    (bit-vectors) such that similar items map to the same or similar hashes
    with a high probability. In other words, it aims to maximize hash collision
    for similar items.

    **Building Models**

    Some hash functions want to build a model based on some training set of
    descriptors. Due to the non-standard nature of algorithm training and model
    building, please refer to the specific implementation for further
    information on whether model training is needed and how it is accomplished.

    """

    def __call__(self, descriptor: np.ndarray) -> np.ndarray:
        return self.get_hash(descriptor)

    @abc.abstractmethod
    def get_hash(self, descriptor: np.ndarray) -> np.ndarray:
        """
        Get the locality-sensitive hash code for the input descriptor.

        :param descriptor: Descriptor vector we should generate the hash of.

        :return: Generated bit-vector as a numpy array of booleans.

        """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['smqtk_indexing',
 'smqtk_indexing.impls',
 'smqtk_indexing.impls.hash_index',
 'smqtk_indexing.impls.lsh_functor',
 'smqtk_indexing.impls.nn_index',
 'smqtk_indexing.interfaces',
 'smqtk_indexing.utils']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19.5,<2.0.0',
 'scipy>=1.5.2,<2.0.0',
 'smqtk-core>=0.18.0',
 'smqtk-dataprovider>=0.16.0',
 'smqtk-descriptors>=0.18.0']

extras_require = \
{':python_version < "3.7"': ['dataclasses>=0.8,<0.9'],
 'faiss': ['faiss-cpu>=1.7.0,<2.0.0'],
 'sklearn': ['scikit-learn>=0.24.1,<0.25.0']}

entry_points = \
{'smqtk_plugins': ['smqtk_indexing.impls.hash_index.linear = '
                   'smqtk_indexing.impls.hash_index.linear',
                   'smqtk_indexing.impls.hash_index.sklearn_balltree = '
                   'smqtk_indexing.impls.hash_index.sklearn_balltree',
                   'smqtk_indexing.impls.lsh_functor.itq = '
                   'smqtk_indexing.impls.lsh_functor.itq',
                   'smqtk_indexing.impls.lsh_functor.simple_rp = '
                   'smqtk_indexing.impls.lsh_functor.simple_rp',
                   'smqtk_indexing.impls.nn_index.faiss = '
                   'smqtk_indexing.impls.nn_index.faiss',
                   'smqtk_indexing.impls.nn_index.flann = '
                   'smqtk_indexing.impls.nn_index.flann',
                   'smqtk_indexing.impls.nn_index.lsh = '
                   'smqtk_indexing.impls.nn_index.lsh',
                   'smqtk_indexing.impls.nn_index.mrpt = '
                   'smqtk_indexing.impls.nn_index.mrpt']}

setup_kwargs = {
    'name': 'smqtk-indexing',
    'version': '0.18.0',
    'description': 'Algorithms, data structures and utilities around computingdescriptor k-nearest-neighbors.',
    'long_description': "# SMQTK - Indexing\n\nThis package provides interfaces and implementations around the\nk-nearest-neighbor algorithm.\n\nThis package defines interfaces and implementations around efficient,\nlarge-scale indexing of descriptor vectors.\nThe sources of such descriptor vectors may come from a multitude of sources,\nsuch as hours of video archives.\nSome provided implementation plugins include [Locality-sensitive Hashing\n(LSH)](https://en.wikipedia.org/wiki/Locality-sensitive_hashing) and FAIR's\n[FAISS] library.\n\n## Documentation\nYou can build the sphinx documentation locally for the most up-to-date\nreference:\n```bash\n# Install dependencies\npoetry install\n# Navigate to the documentation root.\ncd docs\n# Build the docs.\npoetry run make html\n# Open in your favorite browser!\nfirefox _build/html/index.html\n```\n",
    'author': 'Kitware, Inc.',
    'author_email': 'smqtk-developers@kitware.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Kitware/SMQTK-Indexing',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import itertools
from typing import Callable, Iterable, TypeVar


T = TypeVar("T")


def check_empty_iterable(
    iterable: Iterable,
    callback: Callable[[Iterable], T],
    exception_inst: BaseException
) -> T:
    """
    Check that the given iterable is not empty, then call the given callback
    function with the reconstructed iterable when it is not empty.

    :param iterable: Iterable to check.
    :param callback: Function to call with the reconstructed, not-empty
        iterable.
    :param exception_inst: The exception to throw if the iterable is empty

    """
    i = iter(iterable)
    try:
        first = next(i)
    except StopIteration:
        raise exception_inst
    return callback(itertools.chain([first], i))
